
;;; Permission to use, copy, modify, and distribute this software in
;;; object code form for any purpose and without fee is hereby granted.
 
;;; AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS.  
;;; AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF 
;;; MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC.
;;; DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE 
;;; UNINTERRUPTED OR ERROR FREE.

;;; ============ XD2OD.LSP by Peter Heald June 2001 =============

;;; This AutoLISP routine will iterate through a selection set
;;; reading selected xdata, then append that data as
;;; Object Data records. The original xdata is not removed.
;;; It supports all xdata types except braces and binary chunks.
;;; Data is converted into the four OD data types.


(defun C:xd2od ( / xd_list ename ed elist appname rname index
                   lutab applist useit xd_code xd_data tabn
                   odname ss ssiter od_xdata en lu_iter od_col)

  ;;; Keep looping until required xdata sample obtained.

  (while (not xd_list)

    ;;; Get sample entity.

    (while (not (setq ename (car (entsel)))))

    ;;; Get entity data and any associated xdata.

    (setq ed    (entget ename '("*")))
    (setq elist (cdr (assoc -3 ed)))

    ;;; List app names, prompt for name of xdata for conversion.

    (prompt "\nApplication names associated with selected entity:")
    (textscr)
    (if elist
      (progn
        (foreach xchunk elist
          (print (setq appname (car xchunk)))
        )
        (setq rname (getstring (strcat "\nApplication name of xdata to convert? <" appname ">: ")))
        (if (= rname "") (setq rname appname))
        (setq elist (entget ename (list rname)))
      )
      (prompt "\nNo xdata found.")
    )
  
    ;; Check if entity has xdata associated with the entered application name.

    (if (not (setq xd_list (assoc -3 elist)))
      (progn  
        (princ "\nNo Xdata associated with Application Name(s).")
      )
      (setq xd_list (cdr xd_list))      ; Strip -3 from xd_list
    )

  ) ;;; <end of while

  (textscr)
  (setq index 1
        lutab nil
        app_list (car xd_list)
        app_list (cdr app_list)     ;;; Remove app name from xdata
        useit    nil
  )

  ;;; Loop thro each item within xdata 'til end or Finish entered.

  (while (and app_list (/= useit "Finish"))
      (setq xd_code      (car (car app_list))  ;;; Get group code
            xd_data      (cdr (car app_list))  ;;; Get data
      )  

      ;;; Print xdata item and type.
    
      (cond
        ((= 1000 xd_code) (princ "\nCode 1000, string:     ") (princ xd_data))
        ((= 1001 xd_code) (princ "\nCode 1001, REGAPP:     ") (princ xd_data))
        ((= 1003 xd_code) (princ "\nCode 1003, Layer name: ") (princ xd_data))
        ((= 1005 xd_code) (princ "\nCode 1005, Handle:     ") (princ xd_data))
        ((= 1010 xd_code)
          (princ "\nCode 1010, 3 reals:    ")
          (princ (strcat (rtos (car xd_data)) " " (rtos (cadr xd_data)) " " (rtos (caddr xd_data))))
        )
        ((= 1011 xd_code)
          (princ "\nCode 1011, World pos:  ")
          (princ (strcat (rtos (car xd_data)) " " (rtos (cadr xd_data)) " " (rtos (caddr xd_data))))
        )
        ((= 1012 xd_code)
          (princ "\nCode 1012, World disp: ")
          (princ (strcat (rtos (car xd_data)) " " (rtos (cadr xd_data)) " " (rtos (caddr xd_data))))
        )
        ((= 1013 xd_code)
          (princ "\nCode 1013, World dir:  ")
          (princ (strcat (rtos (car xd_data)) " " (rtos (cadr xd_data)) " " (rtos (caddr xd_data))))
        )
        ((= 1040 xd_code) (princ "\nCode 1040, Real No. :  ") (princ (rtos xd_data)))
        ((= 1041 xd_code) (princ "\nCode 1041, Distance:   ") (princ (rtos xd_data)))
        ((= 1042 xd_code) (princ "\nCode 1042, Scale fact: ") (princ (rtos xd_data)))
        ((= 1070 xd_code) (princ "\nCode 1070, 16-bit Int: ") (princ xd_data))
        ((= 1071 xd_code) (princ "\nCode 1071, 32-bit Int: ") (princ (rtos xd_data 2 0)))

      ) ;;; <end of cond

      ;;; Check that data is not a brace or binary chunk.

      (if (and (/= xd_code 1002)(/= xd_code 1004))
        (progn
          (initget 0 "Finish Yes No")

          ;;; Ask whether to convert...

          (setq useit (getkword "\nConvert this to OD ?  Finish/Yes/<No> "))
          (if (= useit "Yes")
            (progn

              ;;; and if so, ask what OD field name to use...

              (while (= (setq tabn (getstring "\nOD field name: " )) ""))

              ;;; ...then append to lookup table.

              (setq lutab (append lutab (list index) (list (cons xd_code tabn))))
            )
          )

          ;;; Increment index - used as placemarker within xdata structure.

          (setq index (1+ index))

        ) ;;; <end of progn

      ) ;;; <end of if

      ;;; Move on to next xdata item.

      (setq app_list (cdr app_list))

  )  ;;; <end of while
  
  ;;; If lookup table exists, xdata was selected for conversion.

  (if lutab
    (progn

      ;;; Get name of OD table to use.

      (setq odname (getstring (strcat "\nName of Object Data table <" rname ">: ")))
      (if (= odname "") (setq odname rname))
      (prompt (strcat "\nSelect entities for xdata conversion to OD table " odname))

      ;;; Keep asking until entities selected.

      (while (not (setq ss (ssget))))
      (setq ssiter (sslength ss))

      ;;; Loop through selection set.

      (while (>= (setq ssiter (1- ssiter)) 0)

        ;;; Get entity name.

        (setq en       (ssname ss ssiter)
              od_xdata nil
        )
   
        ;;; Get xdata associated with appname.
         
        (if (setq xd_list  (assoc -3 (entget en (list rname))))
          (progn
            (textscr)
            (setq xd_list  (cdr xd_list)   ;;; Strip -3
                app_list (car xd_list)
                app_list (cdr app_list)  ;;; Strip app name
                index    1
            )

            ;;; Loop through each xdata item.

            (while app_list

              ;; Get xdata item and group code.

              (setq xd_code      (car (car app_list))   ; Get group code
                    xd_data      (cdr (car app_list))   ; Get data
              )

              ;;; Check that data type is not brace or binary.

              (if  (and (/= xd_code 1002)(/= xd_code 1004))
                (progn

                  ;;; Check if current item index is in lookup table...

                  (if (setq lu_iter (member index lutab))
                    (progn

                      ;;;  ...if it is, we want it.

                      (setq od_col  (cdr (cadr lu_iter)))

                      ;;; Check that data type of item matches that in lookup table...

                      (if (/= (car (cadr lu_iter)) xd_code)

                        ;;; ...if not, produce error.
              
                        (prompt "\nSequence error: mismatched xdata types.")

                        ;;; ...but if ok, add to xdata list.

                        (setq od_xdata (append od_xdata (list (list xd_code (cons od_col xd_data)))))

                      ) ;;; <end if

                    ) ;;; <end of progn

                  ) ;;; <end if

                  ;;; Increment index - this is the important placemarker.

                  (setq index (1+ index))

                ) ;;; <end of progn

              ) ;;; <end if

              (setq app_list (cdr app_list))
 
            ) ;;; <end of while

            ;;; Call add_en_xdata routine to add required xdata to entity.
  
            (add_en_xdata en odname od_xdata)   

          ) ;;; <end of progn

        ) ;;; <end if

      ) ;;; <end of while

    ) ;;; <end of progn

    (prompt "\nNo xdata selected for conversion.")

  ) ;;; <end if

  (princ)

) ;;; <end defun C:xd2od


;;;======================================================
;;; Function add_tabdef to create OD table definition
;;;======================================================

(defun add_tabdef (tabname xdata / tabdef coltype colname defval)

  (prompt (strcat "\nCreating Object Data Table " tabname "."))

  ;;; Loop thro each item (as column) in xdata list.
  
  (foreach col xdata

    ;;; Get column name and xdata item type.

    (setq coltype (car col)
          colname (car (car (cdr col)))
    )

    ;;; Set the OD column type depending on xdata type.
    ;;; Also set the appropriate default value.

    (cond
      ((or (= 1000 coltype)(= 1001 coltype)(= 1003 coltype)(= 1005 coltype))
         (setq coltype "Character" defval "")
      )
      ((or (= 1010 coltype)(= 1011 coltype)(= 1012 coltype)(= 1013 coltype))
         (setq coltype "Point" defval "0.0,0.0,0.0")
      )
      ((or (= 1040 coltype)(= 1041 coltype)(= 1042 coltype))
         (setq coltype "Real" defval 0.0)
      )
      ((or (= 1070 coltype)(= 1071 coltype))
         (setq coltype "Integer" defval 0)
      )
    )

    ;;; Create column def list.

    (setq tabdef (append tabdef 
                   (list
                     (list (cons "colname" colname)
                           (cons "coldesc" colname)
                           (cons "coltype" coltype)
                           (cons "defaultval" defval)
                     )
                   )
                 )
    )

  )  ;;; <end of foreach

  ;;; Add table name etc to column defs.

  (setq tabdef (append (list (cons "tablename" tabname)
                             (cons "tabledesc" "OD from XD")
                       )
                       (list "columns" tabdef)
               )
  )

  ;;; Create OD table def.

  (if (not (ade_oddefinetab tabdef))
    (prompt "\nTable creation failed!\n")
  )

) ;;; <end defun add_tabdef

(princ)


;;;======================================================
;;; Function add_en_xdata to add OD to entity.
;;;======================================================

(defun add_en_xdata (en tabname xdata / recid colname colval)

  ;;; Check if OD table exists...

  (if (not (member tabname (ade_odtablelist)))

    ;;; ...if not, go create it.

    (add_tabdef tabname xdata)
  )

  ;;; Create new OD record in memory.

  (setq recid   (ade_odnewrecord tabname))

  ;;; Loop thro xdata setting record fields.

  (foreach col xdata
    (setq  colname (car (car (cdr col))) 
           colval  (cdr (car (cdr col)))
    )
    (ade_odpresetfield recid colname colval)
  )

  ;;; Attach record to entity.

  (if (ade_odattachrecord en recid)
    (prompt ".")
    (progn
      (prompt "\nFailed to attach new OD record as follows:")
      (print en)
      (princ "Table:") (print tabname)
    )
  )

  ;;; Free memory used by record.

  (ade_odfreerec recid)
  (princ)
)

(princ)


